/**
 * @file
 * Shadow casting object model class implementation.
 *
 * @author Tom "Shamot" Burian
 */
/* ---------------------------------------------------------------------------
 * Authors:      Tom Burian (tburian _AT centrum.cz)            
 * Thanks to:    PCJohn (peciva _AT fit.vutbr.cz)
 * Contributors: 
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 * ---------------------------------------------------------------------------
 */

//----------------------------------------------------------------------------
// INCLUDEs
//----------------------------------------------------------------------------

#include "ShadowManager.h"
#include "ObjectModel.h"

//----------------------------------------------------------------------------
// IMPLEMENTATION
//----------------------------------------------------------------------------

SbVec3f CObjectModel::prevNormal(1.0f, 1.0f, 1.0f);

/**
 * Constructor.
 */
CObjectModel::CObjectModel()
{
  nTriangles = 0;
  modelIfs = NULL;
  modelCoord3 = NULL;
  objectModelRoot = new SoSeparator;
}


/**
 * Destructor.
 */
CObjectModel::~CObjectModel()
{
  objectModelRoot->removeAllChildren();
}


/**
 * Computes the triangle normal vector from its verteces.
 *
 * Counterclockwise orientation.
 *
 * @param p1 point 1
 * @param p2 point 2
 * @param p3 point 3
 *
 * @return face normal vector
 */
SbVec3f CObjectModel::faceNormal(const SbVec3f p1, const SbVec3f p2, const SbVec3f p3)
{
  SbVec3f vector1, vector2;

  vector1 = p2 - p1;
  vector2 = p3 - p2;

  if ( vector1.length() <= .0f ) {
    ONDEBUG(printf("Error: Length of a vector1 is 0. Correcting...\n"));
  }
  if ( vector2.length() <= .0f ) {
    ONDEBUG(printf("Error: Length of a vector2 is 0. Correcting...\n"));
  }

  SbVec3f normal = vector1.cross(vector2);
  if ( normal.length() > .0f ) {
    normal.normalize();
    prevNormal = normal;
  } else {
    ONDEBUG(printf("Error: FaceNormal: Length of the normal is 0. Using previous.\n"));
    ONDEBUG(printf("prevNormal: %.2f %.2f %.2f \n",prevNormal[0], prevNormal[1], prevNormal[2]));
    normal = prevNormal;

//    normal = SbVec3f(0.0f,0.0f,0.0f);
  }

  return normal;
}


/**
 * Creates a SoIndexedFaceSet model of a an object.
 * 
 * This CB is called for every polygon in scene graph.
 *
 * @param userdata Pointer to the object model.
 * @param action   Pointer to callback action.
 * @param v1       Pointer to vertex 1.
 * @param v2       Pointer to vertex 2.
 * @param v3       Pointer to vertex 3.
 */ 
void CObjectModel::TriangleCB(void * userdata,
                       SoCallbackAction * action,
                       const SoPrimitiveVertex * v1,
                       const SoPrimitiveVertex * v2,
                       const SoPrimitiveVertex * v3)
{
  CObjectModel * sc = (CObjectModel *) userdata;
  sc->nTriangles++;

  // get current active material
  SbColor ambient;
  SbColor diffuse;
  SbColor specular; 
  SbColor  emission;
  float shininess;
  float transparency;
  int index = 0;
  action->getMaterial(ambient, diffuse, specular, emission, shininess, transparency, index );
  
/*
  // TODO: grab normal somewhow
  int numnorm = action->getNumNormals();
  faceDetail->getPoint(i)->getCoordinateIndex();
  printf("object model: numnorm: %d \n", numnorm);
*/

  // create model of object's opaque triangles
  if ( transparency == 0.0f )
  {
    // transformation into global coords

	  // triangle vertices
	  const SbVec3f vtx[] = { v1->getPoint(), v2->getPoint(), v3->getPoint() };
	  // tr. matrix
    const SbMatrix mm = action->getModelMatrix();

    // Get tr. coordinates
    SbVec3f vx[3];
    mm.multVecMatrix(vtx[0], vx[0]);
    mm.multVecMatrix(vtx[1], vx[1]);
    mm.multVecMatrix(vtx[2], vx[2]);

    // Add a face to face set
	  sc->modelCoord3->point.setNum(sc->modelCoord3->point.getNum() + 3);
	  sc->modelCoord3->point.setValues(sc->coord3idx, 3, vx);

	  int32_t indices[] = { sc->coord3idx, sc->coord3idx + 1, sc->coord3idx + 2, -1 };
	  sc->coord3idx += 3;

	  int oldsize = sc->modelIfs->coordIndex.getNum();
	  sc->modelIfs->coordIndex.setNum(oldsize + 4);
	  sc->modelIfs->coordIndex.setValues(oldsize, 4, indices);

    // compute normal vector
    // assumes counterclockwise triangle orientation
    SbVec3f normal = faceNormal(vx[0], vx[1], vx[2]);

    // Add normals 1 per face
    int oldsizeN = sc->modelIfs->normalIndex.getNum();
    sc->modelNormal->vector.set1Value(oldsizeN, normal);

    sc->modelIfs->normalIndex.set1Value(oldsizeN, oldsizeN); // ??? TB
	  sc->modelIfs->normalIndex.setNum(oldsizeN + 1);
  }
}


/**
 * Create indexed Face Set model of the object in it's local coordinates.
 *
 * @param object Pointer to the object node.
 */
void CObjectModel::createFaceModel(SoSeparator * object)
{
  ONDEBUG(printf("creating face model...\n"));
  ONDEBUG( printf("  objectModelRoot has %d children\n",objectModelRoot->getNumChildren()) );
  objectModelRoot->removeAllChildren();

  nTriangles = 0;
  coord3idx = 0;

	modelCoord3 = new SoCoordinate3;
	modelCoord3->point.setNum(0);

	modelIfs = new SoIndexedFaceSet;
	modelIfs->coordIndex.setNum(0);
  modelIfs->normalIndex.setNum(0);

	modelNormal = new SoNormal;
	modelNormal->vector.setNum(0);

  SoShapeHints * hints = new SoShapeHints;
  hints->shapeType = SoShapeHintsElement::SOLID;
  hints->vertexOrdering = SoShapeHintsElement::COUNTERCLOCKWISE;
  //hints->vertexOrdering = SoShapeHintsElement::CLOCKWISE;
  
  // call CB action
  SoCallbackAction ca;
  ca.addTriangleCallback(SoShape::getClassTypeId(), TriangleCB, this);
	ca.apply(object);

  // normal for each polygon
  SoNormalBinding *nbind = new SoNormalBinding;
  nbind->value.setValue(SoNormalBinding::PER_FACE_INDEXED);

  objectModelRoot->addChild(modelNormal);
  objectModelRoot->addChild(nbind);
  objectModelRoot->addChild(hints);

	objectModelRoot->addChild(modelCoord3);
	objectModelRoot->addChild(modelIfs);
  
  ONDEBUG(printf("Face model created. Model consists of %d triangles.\n", this->nTriangles));
}


/**
 * Returns the model root node.
 */
SoSeparator * CObjectModel::getFaceModelNode()
{
  return objectModelRoot;
}


/**
 * Returns number of model triangles.
 */
int32_t CObjectModel::getTrianglesCount()
{
  return nTriangles;
}
