/**
 * @file
 * Shadows demonstration application. 
 * Scene creation and user callbacks. 
 *
 * @author Tom "Shamot" Burian
 */
/* ---------------------------------------------------------------------------
 * Authors:      Tom Burian (tburian _AT centrum.cz)            
 * Thanks to:    PCJohn (peciva _AT fit.vutbr.cz)
 * Contributors: 
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 * ---------------------------------------------------------------------------
 */

//----------------------------------------------------------------------------
// INCLUDEs
//----------------------------------------------------------------------------

#include "scene.h"
#include "shadows_demo.h"
#include "cmd_parser.h"
#include "hud.h"


//----------------------------------------------------------------------------
// Program settings
//----------------------------------------------------------------------------

/** input scene (object) file */
char * file = NULL; 
/** show default surface below the object */
int surface = true; 
/** step when moving object */
const float step = 0.5f;

float engFrequency = 0.1f;

struct TShadowSettings settings = {zpass, TRUE, FAR_DEPTH};

int n = 1;

float myzoom = 1.0f;

bool enab = true;
bool shadows = true;
// bool addlight = true;
bool addobject = true;

// global objects, to be able to use them in the demo
SoSeparator * sceneRoot;
SoGroup * lightRoot;
SoGroup * modelRoot;
SoSeparator * objectParent;
SoSeparator * objectParent2;  
SoSeparator * object;
SoSeparator * object2;
SoScale * objectScale;
CShadowManager * shadowManager;
SoEnvironment * env;
SoTransform * transform;
SoTimeCounter *myCounter;

// lights
SoPointLightManip * light1 = NULL;
SoPointLightManip * light2 = NULL;
SoPointLightManip * light3 = NULL;
SoPointLightManip * light4 = NULL;
SoDirectionalLightManip * dLight1 = NULL;
SoSpotLightManip * sLight1 = NULL;

// HUD to display algorithm
CHUD * hud;

//----------------------------------------------------------------------------
// IMPLEMENTATION
//----------------------------------------------------------------------------

/**
 * Checks program parameters and settings.
 *
 * @param argc Number of command line arguments.
 * @param argv Array of command line arguments.
 */
void checkArgs(int argc, char** argv)
{ 

  ONDEBUG(printf("check\n"));
  int i = 1;

  for (i=1; i<argc; i++) {
    
    if ( strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0 ) {
      // Print usage
      printf("Usage: %s [-h|--help] [-f filename] [-fd float] [-d float] [-sv] [-hs] [-zfail]\n"
             "-h, --help        this screen\n"
             "-f filename.iv    load inventor model from file\n"
             "-fd float         depth of shadow volumes\n"             
//             "-fr float         time engine frequency\n"             
             "-hs               hide surface\n"
             "-zfail            use z-fail algorithm (z-pass default)\n",
             argv[0]);
      exit(EXIT_SUCCESS);      

    } else if ( strcmp(argv[i], "-f") == 0 ) {

      // load object from file
      if ( argc >= i+2 ) {
        file = argv[i+1];
        i++;
      } else {
        printf("Parameter -f without file. Specify file to load.\n");
        exit(1);
      }

    } else if ( strcmp(argv[i], "-fd") == 0 ) {

      // far depth
      if ( argc >= i+2 ) {
        char *errp;
        float tmp;
        strtoX(tmp, argv[i+1], &errp);
        if (*(argv[i+1]) != '\0' && *errp == '\0') {
          if ( tmp > 1.0f ) {
            settings.fDepth = tmp;
          } else {
            printf("Warning: Depth of shadow volumes must be > 1.0. Using default %.1f.\n", FAR_DEPTH);
          }          
        }
        i++;
      } else {
        printf("Parameter -fd without value\n");
        exit(1);
      }

    } else if ( strcmp(argv[i], "-fr") == 0 ) {

      // far depth
      if ( argc >= i+2 ) {
        char *errp;
        float tmp;
        strtoX(tmp, argv[i+1], &errp);
        if (*(argv[i+1]) != '\0' && *errp == '\0') {
          if ( tmp > 1.0f ) {
            engFrequency = tmp;
          } else {
            printf("Warning: Time engine frequency  must be > 0.0. Using default %.1f.\n", engFrequency);
          }          
        }
        i++;
      } else {
        printf("Parameter -fr without value\n");
        exit(1);
      }

    } else if ( strcmp(argv[i], "-sv") == 0 ) {
      // create shadows
      settings.dropShadows = FALSE;
    } else if ( strcmp(argv[i], "-hs") == 0 ) {
      // hide surface      
      surface = false;

    } else if ( strcmp(argv[i], "-zfail") == 0 ) {
      // which algorithm to use
      settings.method = zfail;
    } else {
      printf("Warning: Unknown parameter '%s'.\nUsing default settings.\n",argv[i]);
    }
    ONDEBUG(printf("%d : %s\n", i, argv[i]));
  }

   ONDEBUG(printf("check done\n"));

  // Final check
  if (i != argc) {
    printf("Unknown argument %s.\n", argv[i]);
    exit(EXIT_FAILURE);
  }
}


/**
 * Creates scene surface.
 *
 * @param depth Depth of surface. 
 * @param depth Surface color. 
 */
SoSeparator* createSurface(float depth, SbColor color)
{
  // sceneRoot node
  SoSeparator *sceneRoot = new SoSeparator;
  sceneRoot->ref();

  // Material
  SoMaterial *material = new SoMaterial;
  material->ambientColor.setValue(color);
  material->diffuseColor.setValue(color);
  sceneRoot->addChild(material);

  // Translation to set depth
  SoTranslation *trans = new SoTranslation;
  trans->translation.setValue(0.f, 0.f, depth);
  sceneRoot->addChild(trans);

  // Coordinates
  SoCoordinate3 *coords = new SoCoordinate3;
  coords->point.setValues(0, 8, surfaceVertices);
  sceneRoot->addChild(coords);

  for (int i=0; i<1; i++) { // i<7 for other faces
    SoIndexedTriangleStripSet *strip = new SoIndexedTriangleStripSet;
    strip->coordIndex.setNum(1);
    strip->coordIndex.setValues(0, 8, surfaceIndices[i]);
    sceneRoot->addChild(strip);
  }

  sceneRoot->unrefNoDelete();
  return sceneRoot;
}


/**
 * Keyboard callback.
 * Movement control of the first scene object.
 *
 * @param userdata Object translation node.
 * @param keyboardCallback Keyboard callback.
 */
void myKeyPressCallback (void * userdata, SoEventCallback * keyboardCallback)
{
  ONDEBUG( printf("~~~myKeyPressCallback\n") );
  SoTranslation *tr = static_cast<SoTranslation*>(userdata);

  const SoKeyboardEvent* e = static_cast<const SoKeyboardEvent*>(keyboardCallback->getEvent());
  const SbVec3f vector(tr->translation.getValue());

  switch (e->getKey())
  {
    case SoKeyboardEvent::UP_ARROW: // move object on y+ axis 
      tr->translation.setValue(vector[0], vector[1]+step, vector[2]);
      keyboardCallback->setHandled();
    break;

    case SoKeyboardEvent::DOWN_ARROW: // move object on y- axis 
      tr->translation.setValue(vector[0], vector[1]-step, vector[2]);
      keyboardCallback->setHandled();
    break;

    case SoKeyboardEvent::LEFT_ARROW: // move object on x- axis       
      tr->translation.setValue(vector[0]-step, vector[1], vector[2]);
      keyboardCallback->setHandled();
    break;

    case SoKeyboardEvent::RIGHT_ARROW: // move object on x+ axis 
      tr->translation.setValue(vector[0]+step, vector[1], vector[2]);
      keyboardCallback->setHandled();
    break;

    case SoKeyboardEvent::PAGE_DOWN: // move object on z- axis 
      tr->translation.setValue(vector[0], vector[1], vector[2]-step);
      keyboardCallback->setHandled();
    break;

    case SoKeyboardEvent::PAGE_UP: // move object on z+ axis 
      tr->translation.setValue(vector[0], vector[1], vector[2]+step);
      keyboardCallback->setHandled();
    break;

    default:
//      keyboardCallback->setHandled();
    break;
  }
  ONDEBUG( printf("~~~myKeyPressCallback done\n") );
}


void myObjectChangeCallback (void * userdata, SoEventCallback * keyboardCallback)
{
  ONDEBUG( printf("~~~myObjectChangeCallback\n") );
  SoSeparator *obj = static_cast<SoSeparator*>(userdata);

  const SoEvent * event = keyboardCallback->getEvent();
  const SoKeyboardEvent* e = static_cast<const SoKeyboardEvent*>(keyboardCallback->getEvent());

  SoSphere * sphere = new SoSphere;
  SoTranslation *objectTranslation = new SoTranslation;
  objectTranslation->translation.setValue(0.f, 0.f, (-1)*n*5.f);  

  if ( !keyboardCallback->isHandled() ) {
    switch (e->getKey())
    {
      case SoKeyboardEvent::SPACE: // zmena tvaru prvniho objektu, prida se koule
        if (e->isKeyPressEvent(event, SoKeyboardEvent::SPACE) ) {
          ONDEBUG( printf("\nSpace\n") );
          sphere->radius = 3;
          obj->addChild(objectTranslation);
          obj->addChild(sphere);
          keyboardCallback->setHandled();
        }
      break;
    }
    // keyboardCallback->setHandled();
  }
  ONDEBUG( printf("~~~myObjectChangeCallback done\n") );
}



/**
 * Keyboard callback3.
 * Removing objects test.
 *
 * @param userdata SM.
 * @param keyboardCallback Keyboard callback.
 */
void myKeyPressCallback3 (void * userdata, SoEventCallback * keyboardCallback)
{
  ONDEBUG( printf("~~~myKeyPressCallback3\n") );
  CShadowManager * sm = static_cast<CShadowManager*>(userdata);

  int cnt = -1;
  const SoEvent * event = keyboardCallback->getEvent();
  const SoKeyboardEvent* e = static_cast<const SoKeyboardEvent*>(keyboardCallback->getEvent());
  switch (e->getKey())
  {
    case SoKeyboardEvent::PAD_SUBTRACT: // remove first object
      if (e->isKeyPressEvent(event, SoKeyboardEvent::PAD_SUBTRACT) ) {
        ONDEBUG( printf("PAD_SUBTRACT:\n") );
        sm->removeObject(object); // tento tam bude vzdy, jen pro test. ucely
        keyboardCallback->setHandled();
      }
    break;

    case SoKeyboardEvent::PAD_ADD: // add first object
      if (e->isKeyPressEvent(event, SoKeyboardEvent::PAD_ADD) ) {        
        ONDEBUG( printf("PAD_ADD:\n") );
        sm->addObject(object);         
        keyboardCallback->setHandled();
      }
    break;


    case SoKeyboardEvent::F1: // add zoom in scale to the object
      if (e->isKeyPressEvent(event, SoKeyboardEvent::F1) ) {         
        ONDEBUG( printf("DISABLE/ENABLE\n") );
        if (shadows)
        {
          sm->setSceneType(CShadowManager::SCENE);
          shadows = !shadows;
          hud->refresh("Scene without shadows");
        }
        else
        {
          sm->setSceneType(CShadowManager::SHADOWS);
          shadows = !shadows;
          if ( sm->readSettings()->method == zpass )
            hud->refresh("Algorithm: z-pass");
          else
            hud->refresh("Algorithm: z-fail");
        }
        keyboardCallback->setHandled();
      } 
    break;


    case SoKeyboardEvent::F2: // set z-pass or z-fail on fly
      if (e->isKeyPressEvent(event, SoKeyboardEvent::F2) ) {        
        if ( sm->readSettings()->method == zpass )
        {
          ONDEBUG( printf("SM is using z-zfail:\n") );
          sm->setMethod(zfail);
          if ( shadows ) 
            hud->refresh("Algorithm: z-fail");
        }
        else
        {
          ONDEBUG( printf("SM is using z-pass:\n") );
          sm->setMethod(zpass);
          if ( shadows ) 
            hud->refresh("Algorithm: z-pass");
        }
        keyboardCallback->setHandled();
      }
    break;


    case SoKeyboardEvent::F3: // add zoom in scale to the object
      if (e->isKeyPressEvent(event, SoKeyboardEvent::F3) ) {         
        ONDEBUG( printf("DISABLE/ENABLE\n") );
        if (enab)
        {
          sm->disableSensors();        
          enab = !enab;
        }
        else
        {
          sm->enableSensors();        
          enab = !enab;
        }
        keyboardCallback->setHandled();
      }
    break;


    case SoKeyboardEvent::F5: // add zoom in scale to the object
      if (e->isKeyPressEvent(event, SoKeyboardEvent::F5) ) {         
        ONDEBUG( printf("SCALE: zoom in\n") );
        object->ref();
        objectParent->removeChild(object);

        SoScale * scale = new SoScale;
        scale->scaleFactor.setValue(2.0f,2.0f,2.0f);
        objectParent->addChild(scale);
        objectParent->addChild(object);
        object->unref();
        shadowManager->createObjectsShadows(object); // force
        keyboardCallback->setHandled();
      }
    break;

    case SoKeyboardEvent::F6: // add zoom out scale to the object
      if (e->isKeyPressEvent(event, SoKeyboardEvent::F6) ) {
        ONDEBUG( printf("SCALE: zoom out\n") );
        object->ref();
        objectParent->removeChild(object);

        SoScale * scale = new SoScale;
        scale->scaleFactor.setValue(0.5f,0.5f,0.5f);
        objectParent->addChild(scale);
        objectParent->addChild(object);
        object->unref();
        shadowManager->createObjectsShadows(object);  // force
      }
    break;

    case SoKeyboardEvent::F7: // add zoom in scale to the object
      if (e->isKeyPressEvent(event, SoKeyboardEvent::F7) ) {         
        ONDEBUG( printf("SCALE node: zoom in\n") );
        myzoom += 1.0f;
        objectScale->scaleFactor.setValue(myzoom,myzoom,myzoom);
      }
    break;

    case SoKeyboardEvent::F8: // add zoom out scale to the object
      if (e->isKeyPressEvent(event, SoKeyboardEvent::F8) ) {
        ONDEBUG( printf("SCALE node: zoom out\n") );
        myzoom -= 1.0f;
        objectScale->scaleFactor.setValue(myzoom,myzoom,myzoom);
      }
    break;

    case SoKeyboardEvent::F9: // on / off light1 toggle
      if (e->isKeyPressEvent(event, SoKeyboardEvent::F9) ) {
        ONDEBUG( printf("on / off light1 toggle\n") );        
        if ( light1 != NULL )
        {
          light1->on.enableNotify(FALSE);
          SbBool s = light1->on.getValue();
          light1->on.enableNotify(TRUE);
          light1->on.setValue( ! s );
//          ONDEBUG( printf("light1 status: %d\n", light1->on.getValue() ) );
        }
      }
    break;

    case SoKeyboardEvent::F11: // increase ambient light intensity
      if (e->isKeyPressEvent(event, SoKeyboardEvent::F11) ) {
        ONDEBUG( printf("increase ambient\n") );
        env->ambientIntensity.setValue(env->ambientIntensity.getValue()+0.1f);
      }
    break;

    case SoKeyboardEvent::F12: // decrease ambient light intensity
      if (e->isKeyPressEvent(event, SoKeyboardEvent::F12) ) {
        ONDEBUG( printf("decrease ambient\n") );
        env->ambientIntensity.setValue(env->ambientIntensity.getValue()-0.1f);
      }
    break;


//////////////////////////
    // add / remove lights on fly

    case SoKeyboardEvent::NUMBER_1:
      ONDEBUG( printf("NUMBER_1 event\n") );
      if (e->isKeyPressEvent(event, SoKeyboardEvent::NUMBER_1) ) {
        
        if ( light1 == NULL )
        {
          ONDEBUG( printf("ADD light1: \n") );
          light1 = new SoPointLightManip;
          light1->location.setValue(-30.0f, -10.0f, 0.0f);           
          light1->intensity.setValue(1.f);
          light1->color.setValue(1.0f, 1.0f, 1.0f); 
          light1->on = TRUE;
          light1->setName("light1");
          lightRoot->addChild(light1);
          shadowManager->addLight(light1);
        }
        else
        {
          ONDEBUG( printf("REMOVE light1: \n") );
          shadowManager->removeLight(light1);
          lightRoot->removeChild(light1);
          light1 = NULL;
        }
      }
    break;

    case SoKeyboardEvent::NUMBER_2:
      ONDEBUG( printf("NUMBER_2 event\n") );
      if (e->isKeyPressEvent(event, SoKeyboardEvent::NUMBER_2) ) {
        
        if ( light2 == NULL )
        {
          ONDEBUG( printf("ADD light2: \n") );
          light2 = new SoPointLightManip;
          light2->location.setValue(-10.0f, -10.0f, 0.0f);           
          light2->intensity.setValue(1.f);
          light2->color.setValue(1.0f, 0.0f, 0.0f); 
          light2->on = TRUE;
          light2->setName("light2");
          lightRoot->addChild(light2);
          shadowManager->addLight(light2);
        }
        else
        {
          ONDEBUG( printf("REMOVE light2: \n") );
          shadowManager->removeLight(light2);
          lightRoot->removeChild(light2);
          light2 = NULL;
        }
      }
    break;

    case SoKeyboardEvent::NUMBER_3:
      ONDEBUG( printf("NUMBER_3 event\n") );
      if (e->isKeyPressEvent(event, SoKeyboardEvent::NUMBER_3) ) {
        
        if ( light3 == NULL )
        {
          ONDEBUG( printf("ADD light3: \n") );
          light3 = new SoPointLightManip;
          light3->location.setValue(10.0f, -10.0f, 0.0f);           
          light3->intensity.setValue(1.f);
          light3->color.setValue(0.0f, 1.0f, 0.0f); 
          light3->on = TRUE;
          light3->setName("light3");
          lightRoot->addChild(light3);
          shadowManager->addLight(light3);
        }
        else
        {
          ONDEBUG( printf("REMOVE light3: \n") );
          shadowManager->removeLight(light3);
          lightRoot->removeChild(light3);
          light3 = NULL;
        }
      }
    break;

    case SoKeyboardEvent::NUMBER_4:
      ONDEBUG( printf("NUMBER_4 event\n") );
      if (e->isKeyPressEvent(event, SoKeyboardEvent::NUMBER_4) ) {
        
        if ( light4 == NULL )
        {
          ONDEBUG( printf("ADD light4: \n") );
          light4 = new SoPointLightManip;
          light4->location.setValue(30.0f, -10.0f, 0.0f);           
          light4->intensity.setValue(1.f);
          light4->color.setValue(0.0f, 0.0f, 1.0f); 
          light4->on = TRUE;
          light4->setName("light4");
          lightRoot->addChild(light4);
          shadowManager->addLight(light4);
        }
        else
        {
          ONDEBUG( printf("REMOVE light4: \n") );
          shadowManager->removeLight(light4);
          lightRoot->removeChild(light4);
          light4 = NULL;
        }
      }
    break;

    case SoKeyboardEvent::NUMBER_5:
      ONDEBUG( printf("NUMBER_5 event\n") );
      if (e->isKeyPressEvent(event, SoKeyboardEvent::NUMBER_5) ) {
        
        if ( dLight1 == NULL )
        {
          ONDEBUG( printf("ADD light5: \n") );
          dLight1 = new SoDirectionalLightManip;          
          dLight1->color.setValue(1.0f, 1.0f, 1.0f); 
          dLight1->intensity.setValue(1.f);
          dLight1->on = TRUE;
          dLight1->setName("dLight1");
          lightRoot->addChild(dLight1);
          shadowManager->addLight(dLight1);
        }
        else
        {
          ONDEBUG( printf("REMOVE dLight1: \n") );
          shadowManager->removeLight(dLight1);
          lightRoot->removeChild(dLight1);
          dLight1 = NULL;
        }
      }
    break;


    case SoKeyboardEvent::NUMBER_6:
      ONDEBUG( printf("NUMBER_6 event\n") );
      if (e->isKeyPressEvent(event, SoKeyboardEvent::NUMBER_6) ) {
        
        if ( sLight1 == NULL )
        {
          ONDEBUG( printf("ADD light6: \n") );
          sLight1 = new SoSpotLightManip;          
          sLight1->color.setValue(1.0f, 1.0f, 1.0f); 
          sLight1->intensity.setValue(1.f);
          sLight1->location.setValue(0.0f, -10.0f, 15.0f);
          sLight1->on = TRUE;
          sLight1->setName("sLight1");
          lightRoot->addChild(sLight1);
          shadowManager->addLight(sLight1);
        }
        else
        {
          ONDEBUG( printf("REMOVE sLight1: \n") );
          shadowManager->removeLight(sLight1);
          lightRoot->removeChild(sLight1);
          sLight1 = NULL;
        }
      }
    break;



/////////////////////////

    case SoKeyboardEvent::BRACKETLEFT: // increase light1 intensity
      if (e->isKeyPressEvent(event, SoKeyboardEvent::BRACKETLEFT) ) {
        ONDEBUG( printf("increase light1\n") );
        light1->intensity.setValue(light1->intensity.getValue()+0.1f);
        keyboardCallback->setHandled();
      }
    break;

    case SoKeyboardEvent::BRACKETRIGHT: // decrease ambient light intensity
      if (e->isKeyPressEvent(event, SoKeyboardEvent::BRACKETRIGHT) ) {
        ONDEBUG( printf("decrease light1\n") );
        light1->intensity.setValue(light1->intensity.getValue()-0.1f);
        keyboardCallback->setHandled();
      }
    break;


    case SoKeyboardEvent::X: // save sceneRoot into a file "sceneRoot.iv"
      if (e->isKeyPressEvent(event, SoKeyboardEvent::X) ) {
        ONDEBUG( printf("Saving sceneRoot.\n") );
        SoOutput output;
        output.openFile("SceneRoot.iv");
        SoWriteAction * writer = new SoWriteAction(&output);
        writer->apply(sceneRoot);
        delete writer;
      }
    break;


    case SoKeyboardEvent::V: // save shadow volumes scene graph
      if (e->isKeyPressEvent(event, SoKeyboardEvent::V) ) {
        ONDEBUG( printf("save shadow scene\n") );
        sm->export("ShadowVolumes.iv");
      }
    break;

/*
    case SoKeyboardEvent::A: // slide object1
      if (e->isKeyPressEvent(event, SoKeyboardEvent::A) ) {
        ONDEBUG( printf("slide object along X\n") );
        if ( myCounter->on.getValue() == FALSE )
          myCounter->on = TRUE;
        else
          myCounter->on = FALSE;

      }
    break;
*/


/*
    case SoKeyboardEvent::W: // on / off light1 toggle
      if (e->isKeyPressEvent(event, SoKeyboardEvent::W) ) {
        ONDEBUG( printf("on / off light2 toggle\n") );        
        if ( light2 != NULL )
        {
          light2->on.enableNotify(FALSE);
          SbBool s = light2->on.getValue();
          light2->on.enableNotify(TRUE);
          light2->on.setValue( ! s );
//          ONDEBUG( printf("light1 status: %d\n", light1->on.getValue() ) );
        }
      }
    break;
*/    

    case SoKeyboardEvent::O: // add / remove object on fly
      ONDEBUG( printf("O -  event\n") );
      if (e->isKeyPressEvent(event, SoKeyboardEvent::O) ) {
        ONDEBUG( printf("add / del object2:O \n") );
        if ( addobject )
        {
          ONDEBUG( printf("add / del object2: ADD \n") );

          objectParent2 = new SoSeparator; 
          objectParent2->setName("objectParent2");

          // druhy objekt
          char * filename2 = "simple_ship.iv"; // fighter
          SoInput in;
          if (!in.openFile(filename2)) {
            printf("Can not open input file '%s'.\n", filename2);
            exit(-1);
          }
          object2 = NULL;
          object2 = SoDB::readAll(&in);
          if (object2 == NULL) {
            printf("Can not read input file '%s'.\n", filename2);
            exit(-1);
          }

          // Move the object
          SoTranslation *objectTranslation2 = new SoTranslation;
          objectTranslation2->translation.setValue(30.f, 10.f, 0.0f);
          objectParent2->addChild(objectTranslation2);
 
          // Rotate object
          SoRotationXYZ *objectRotationX2 = new SoRotationXYZ;
          objectRotationX2->axis = SoRotationXYZ::X;
          objectRotationX2->angle.setValue(float(3.1415926/2));
          objectParent2->addChild(objectRotationX2);

          objectParent2->addChild(object2);
          modelRoot->addChild(objectParent2);
          shadowManager->addObject(object2);
          addobject = !addobject;

          ONDEBUG( 
            printf("a new object added\n");
            printf("object2 ref cnt %d\n",object2->getRefCount());
            printf("objectParent2 ref cnt %d\n",objectParent2->getRefCount());
          );
        }
        else
        {
          ONDEBUG( 
            printf("add / del object2: REMOVE \n"); 
            printf("object2 ref cnt %d\n",object2->getRefCount());
            printf("objectParent2 ref cnt %d\n",objectParent2->getRefCount());
            printf("deleting ---------\n");
          );

          shadowManager->removeObject(object2);
          
          ONDEBUG( 
            printf("object2 ref cnt %d\n",object2->getRefCount());
            printf("objectParent2 ref cnt %d\n",objectParent2->getRefCount());
          );

          modelRoot->removeChild(objectParent2);          
          object2 = NULL;
          objectParent2 = NULL;
          addobject = !addobject;
        }
      }
    break;

    default:
//      keyboardCallback->setHandled(); // poslat dal...
    break;
  }
  keyboardCallback->setHandled();
  ONDEBUG( printf("~~~myKeyPressCallback3 done\n") );
}



/**
 * Creates the complete scene with objects.
 *
 * @param argc Args count.
 * @param argv Args.
 */
SoSeparator* createScene(int argc, char **argv)
{
  checkArgs(argc, argv);

  printf("\n%s %s\n", APPNAME, VERSION);
  printf("After ESC key:\n");
  printf("F1 ............. toggle shadows on/off \n");
  printf("F2 ............. toggle z-fail/z-pass algorithm \n");  
  printf("F3 ............. toggle enable/disable object sensors \n");
  printf("{/} ............ increase/decrease light1 intensity \n");
  printf("F5/F6 .......... double/half object size\n");
  printf("F7/F8 .......... object zoom ++/-- \n");
  printf("1-4 ............ add/remove new point light \n");
  printf("5 .............. add/remove new directional light \n");
  printf("6 .............. add/remove new spot light \n");
  printf("F9 ............. toggle light1 on/off \n");
  printf("F11/F12 ........ increase/decrease ambient light intensity \n");
  printf("ARROWS ......... move object1 along x and y axes \n");
  printf("PAGE UP/DOWN ... move object1 along z axe \n");
  printf("key O .......... add/remove the second object \n");
//  printf("key A .......... start/stop sliding the first object along X axe\n");
  printf("key V .......... export shadow volumes into ShadowVolumes.iv \n");
  printf("key X .......... export scene into SceneRoot.iv \n");
  printf("Use manipulators to move the lights.\n\n");

  // One root to rule them all :)
  SoSeparator * topRoot = new SoSeparator;
  topRoot->setName("topRoot");
  topRoot->ref();

  // main node of the shadow scene graph
  SoSeparator * shadowRoot = new SoSeparator;
  shadowRoot->setName("shadowRoot");

  // user's scene definition

  // main node of the scene
  sceneRoot = new SoSeparator;
  sceneRoot->setName("sceneRoot");

  // environment - ambient light
  env = new SoEnvironment;
  env->ambientColor.setValue(SbColor(1.0f, 1.0f, 1.0f));
  env->ambientIntensity.setValue(0.2f);
  sceneRoot->addChild(env);

  // node to collect all lights lighting the scene
  lightRoot = new SoGroup;
  lightRoot->setName("lightRoot");
  sceneRoot->addChild(lightRoot);

  // node to collect all models
  modelRoot = new SoGroup;
  modelRoot->setName("modelRoot");
  sceneRoot->addChild(modelRoot);

  // Light 1
  light1 = new SoPointLightManip;
  light1->color.setValue(1.0f, 1.0f, 1.0f);
  light1->location.setValue(-30.0f, -10.0f, 0.0f); 
  light1->intensity.setValue(1.0f);
  light1->on = TRUE;
  light1->setName("light1");
  lightRoot->addChild(light1);

  // Camera
  SoPerspectiveCamera * camera = new SoPerspectiveCamera;
  camera->position = SbVec3f(0.f, -80.f, 20.f);
  camera->pointAt(SbVec3f(0, 0, -10));

/*
  // Near & Far planes
  camera->nearDistance = 1.f;  // default is  1.0
  camera->farDistance = 1000.f; // default is 10.0
*/

  // First, create a parent node for a new object
//  SoSeparator * objectParent = new SoSeparator;  
  objectParent = new SoSeparator; 
  objectParent->setName("objectParent");

  // time engine 

  transform = new SoTransform();
  objectParent->addChild(transform);

  // Make the Y translation value change over time.
  myCounter = new SoTimeCounter;
  // set 100 ticks per second
  myCounter->min = 0;
  myCounter->max = 100;  
  myCounter->frequency = engFrequency; // 1.0f
  myCounter->on = FALSE;
  //SoElapsedTime *myCounter = new SoElapsedTime;
  SoComposeVec3f *slideDistance = new SoComposeVec3f;
  slideDistance->x.connectFrom( &myCounter->output);
   // slideDistance->y.connectFrom(&myCounter->timeOut);
  transform->translation.connectFrom(&slideDistance->vector); // slideTranslation
  // time engine end

  // Move the object
  SoTranslation *objectTranslation = new SoTranslation;
  objectTranslation->translation.setValue(0.f, 10.f, -25.0f);
  //special for stadium model
  // objectTranslation->translation.setValue(1900.f, -400.f, -400.0f);
  objectParent->addChild(objectTranslation);

  // scale test
  objectScale = new SoScale;
  objectScale->scaleFactor.setValue(myzoom,myzoom,myzoom);
  objectParent->addChild(objectScale);
 
  // Rotate object
  SoRotationXYZ *objectRotationX = new SoRotationXYZ;
  objectRotationX->axis = SoRotationXYZ::X;
  objectRotationX->angle.setValue(float(3.1415926/2));
  objectParent->addChild(objectRotationX);

  // Load object
  char * filename = NULL;
  if ( file != NULL ) filename = file;
  else {
    filename = "simple_ship.iv"; // simple_ship
//    filename = "fighter.iv";
//    filename = "menhir.iv";
    //filename = "stonehenge.iv";
    //filename = "stadium3.iv";
  }

  SoInput in;
  if (!in.openFile(filename)) {
    printf("Can not open input file '%s'.\n", filename);
    exit(-1);
  }

  //SoSeparator *object = SoDB::readAll(&in);
  object = SoDB::readAll(&in);
  object->setName("object1");
  if (object == NULL) {
    printf("Can not read input file '%s'.\n", filename);
    exit(-1);
  }
  objectParent->addChild(object);

//----------------------------------------------------------------------------
// Scene definition
//---------------------------------------------------------------------------- 

  if ( surface )
    sceneRoot->addChild(createSurface(SURFACE_DEPTH, SbColor(0.7f, 0.6f, 0.5f)));

  modelRoot->addChild(objectParent);

  // Manager

  // create Shadow manager
  shadowManager = new CShadowManager(sceneRoot, NULL); // env = NULL - problem ???  

  shadowManager->addLight(light1);
  shadowManager->addObject(object);
  // set up SM
  shadowManager->setManager(&settings);
//  shadowManager->setSceneType(CShadowManager::SHADOWS); // SCENE
//  shadowManager->setMethod(autoset);
//  shadowManager->printSettings();  

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
 
  // Keyboard Callbacks (first object movement by default)
  SoEventCallback * myEventCallback = new SoEventCallback;
  myEventCallback->addEventCallback(SoKeyboardEvent::getClassTypeId(), myKeyPressCallback, objectTranslation);
  sceneRoot->addChild(myEventCallback);

  // Keyboard Callbacks (first object shape change by default)
  SoEventCallback * myEventCallbackG = new SoEventCallback;
  myEventCallbackG->addEventCallback(SoKeyboardEvent::getClassTypeId(), myObjectChangeCallback, object); // objectParent
  sceneRoot->addChild(myEventCallbackG);

  // Keyboard Callbacks 3 (test of removing objects, lights, zoom)
  SoEventCallback * myEventCallback3 = new SoEventCallback;
  myEventCallback3->addEventCallback(SoKeyboardEvent::getClassTypeId(), myKeyPressCallback3, shadowManager);
  sceneRoot->addChild(myEventCallback3);

  topRoot->addChild(camera);
  topRoot->addChild(shadowManager->getShadowSceneRoot());

  char * alg = "";
  if ( *shadowManager->getAlgorithm() == zfailAlg )
    alg = "Algorithm: z-fail";
  else
    alg = "Algorithm: z-pass";


  // with HUD directional light manip do not work !!!
  hud = new CHUD;  
  hud->init(sceneRoot);
  hud->refresh(alg);


  // test
  //delete shadowManager;

/*  
  // Debug: write out scene graph
  SoOutput output;
  output.openFile("vystup.iv");
  SoWriteAction * writer = new SoWriteAction(&output);
  writer->apply(topRoot);
  delete writer;
*/
  

  topRoot->unrefNoDelete();

  return topRoot;
}
//////////////////////////////////////////////////////////////////////////////

